
#ifdef WIN32
#include <Windows.h>
#endif

#include <GL/glut.h>

#include "display.h"
#include "world.h"
#include "vector.h"


extern world *currentWorld;

vector2i windowPosition(100,100);
vector2i windowDimensions(800,600);
vector2f worldDrawPosition(0.1f,0.1f);
vector2f worldDrawDimensions(0.8f,0.8f);

int screenSizeToPixelSize(const float& f)
{
	return (int)(((float)(windowDimensions.x+windowDimensions.y)/2.0f)*f);
}

vector2i screenPosToPixelPos(const vector2f& pos)
{
	return vector2i((int)(pos.x*windowDimensions.x), (int)(pos.y*windowDimensions.y));
}

vector2f pixelPosToScreenPos(const vector2i& pos)
{
	return vector2f((float)pos.x/windowDimensions.x,(float)pos.y/windowDimensions.y);
}

extern bool displayInit(int argc,char** argv)
{
	glutInit( &argc, argv );

	glutInitDisplayMode( GLUT_DEPTH | GLUT_RGBA | GLUT_DOUBLE );
	glutInitWindowPosition(windowPosition.x,windowPosition.y);
	glutInitWindowSize(windowDimensions.x,windowDimensions.y);
	glutCreateWindow("experiment01");

	glutDisplayFunc(draw);
	glutReshapeFunc(resize);

	return true;
}
	
void drawWorld()
{
	unsigned int i,j;
	vector2i drawPos = screenPosToPixelPos(worldDrawPosition);
	vector2i drawDim = screenPosToPixelPos(worldDrawDimensions);
	int lineWidth = 0;

	glViewport(drawPos.x,drawPos.y,drawDim.x,drawDim.y);
			
	
	glColor3f(0.2f,0.7f,0.4f);
	glBegin(GL_QUADS);
			glVertex3f(0,0,0);
			glVertex3f(1,0,0);
			glVertex3f(1,1,0);
			glVertex3f(0,1,0);
	glEnd();

	for(i=0;i<currentWorld->length;i++)
	{
		for(j=0;j<currentWorld->width;j++)
		{
			if(currentWorld->grid[i][j]->hasFunction(TblockedTile))
			{
				float n = (float)j/currentWorld->length;
				float e = (float)(i+1)/currentWorld->length;
				float s = (float)(j+1)/currentWorld->width;
				float w = (float)i/currentWorld->width;

				glColor3f(0.5f,0.5f,0.5f);

				glBegin(GL_QUADS);
					glVertex3f(s,w,0);
					glVertex3f(n,w,0);
					glVertex3f(n,e,0);
					glVertex3f(s,e,0);
				glEnd();
			} else if(currentWorld->grid[i][j]->hasFunction(TopenTile))
			{
				float n = (float)j/currentWorld->length;
				float e = (float)(i+1)/currentWorld->length;
				float s = (float)(j+1)/currentWorld->width;
				float w = (float)i/currentWorld->width;
				
				if(!currentWorld->isTileFree(vector2i(i,j)))
				{
					glColor3f(1.0f,0.0f,0.0f);

					glBegin(GL_QUADS);
						glVertex3f(s,w,0);
						glVertex3f(n,w,0);
						glVertex3f(n,e,0);
						glVertex3f(s,e,0);
					glEnd();
				} else if(currentWorld->doesTileContainFood(vector2i(i,j)))
				{
					glColor3f(0.1f,0.9f,0.9f);

					glBegin(GL_QUADS);
						glVertex3f(s,w,0);
						glVertex3f(n,w,0);
						glVertex3f(n,e,0);
						glVertex3f(s,e,0);
					glEnd();
				}
			}
		}
	}

	glColor3f(1.0f,1.0f,1.0f);

	lineWidth = screenSizeToPixelSize(0.002f);
	glLineWidth(lineWidth?lineWidth:1);
	glBegin(GL_LINES);
		for(i=0;i<currentWorld->length;i++)
		{
			glVertex3f(0,(float)i/currentWorld->length,0);
			glVertex3f(1,(float)i/currentWorld->length,0);
		}

		for(j=0;j<currentWorld->width;j++)
		{
			glVertex3f((float)j/currentWorld->width,0,0);
			glVertex3f((float)j/currentWorld->width,1,0);
		}
	glEnd();


	glColor3f(1.0f,1.0f,1.0f);
	
	lineWidth = screenSizeToPixelSize(0.01f);
	glLineWidth(lineWidth?lineWidth:1);
	glBegin(GL_LINE_LOOP);
			glVertex3f(0,0,0);
			glVertex3f(1,0,0);
			glVertex3f(1,1,0);
			glVertex3f(0,1,0);
	glEnd();
}

void draw()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glOrtho(0,1,0,1,-1,1);

	glViewport(worldDrawPosition.x,worldDrawPosition.y,worldDrawDimensions.x,worldDrawDimensions.y);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	drawWorld();

	glFlush();
	glutSwapBuffers();
}

void resize(int x, int y)
{
	windowDimensions.x = x;
	windowDimensions.y = y;
}
