
#include <iostream>

#include "debug.h"

#include "vector.h"
#include "entity.h"
#include "world.h"
#include "tile.h"

extern world* currentWorld;

unsigned int entity::newProgSeed = 0;

entity::entity(unsigned int energy_)
{
	energy = energy_;
	pc = 0;
	age = 0;
	prog = new unsigned int[PROG_SIZE];

	srand(newProgSeed++);
	for(int i=0;i<PROG_SIZE;i++)
	{
		prog[i] = rand();
	}

}

entity::entity(unsigned int energy_, unsigned int *prog_)
{
	energy = energy_;
	pc = 0;
	age = 0;
	prog = prog_;
}

entity::~entity()
{
	delete [] prog;
}

void entity::move(char dir)
{
	vector2i dest;

	switch(dir)
	{
	case 'n':
		dest.x = pos.x;
		dest.y = pos.y+1;
		break;
	case 'e':
		dest.x = pos.x+1;
		dest.y = pos.y;
		break;
	case 's':
		dest.x = pos.x;
		dest.y = pos.y-1;
		break;
	case 'w':
		dest.x = pos.x-1;
		dest.y = pos.y;
		break;
	}

	currentWorld->moveEntity(*this, dest);
}

void entity::update()
{
	age++;
	energy--;

	switch(prog[pc]%6)
	{
	case 0:
		BNE();
		break;
	case 1:
		MOV();
		break;
	case 2:
		ADD();
		break;
	case 3:
		SUB();
		break;
	case 4:
		DAT();
		break;
	case 5:
		EXE();
		break;
	}

	pc%=PROG_SIZE;

	if(!energy)
	{
		currentWorld->killEntity(this);
	}
}

void entity::BNE()
{
	// BNE,[test],[address]

	unsigned int a = (pc+1<PROG_SIZE)?pc+1:0;
	unsigned int b = (pc+2<PROG_SIZE)?pc+2:1;
	
	if(prog[a]%2)
	{
		pc = prog[b]%PROG_SIZE;
	} else
	{
		pc+=3;
	}
}

void entity::MOV()
{
	// MOV,[from],[to]

	unsigned int a = (pc+1)%PROG_SIZE;
	unsigned int b = (pc+2)%PROG_SIZE;

	prog[b]=prog[a];

	pc+=3;
}

void entity::ADD()
{
	// ADD,[operand],[operand],[result]

	unsigned int a = (pc+1)%PROG_SIZE;
	unsigned int b = (pc+2)%PROG_SIZE;
	unsigned int c = (pc+3)%PROG_SIZE;

	prog[c] = prog[a]+prog[b];

	pc+=4;

}

void entity::SUB()
{
	// SUB,[operand],[operand],[result]

	unsigned int a = (pc+1)%PROG_SIZE;
	unsigned int b = (pc+2)%PROG_SIZE;
	unsigned int c = (pc+3)%PROG_SIZE;

	prog[c] = prog[a]-prog[b];

	pc+=4;

}

void entity::DAT()
{
	unsigned int a = (pc+1)%PROG_SIZE;

	unsigned int datType = prog[a]%2;
	if(datType == 0)
	{
		// DAT,[type],[direction],[result]
		unsigned int b = (pc+2)%PROG_SIZE;
		unsigned int c = (pc+3)%PROG_SIZE;
		switch(prog[b]%4)
		{
		case 0:
			prog[c] = currentWorld->isTileFree(vector2i(pos.x,pos.y+1));
			break;
		case 1:
			prog[c] = currentWorld->isTileFree(vector2i(pos.x,pos.y+1));
			break;
		case 2:
			prog[c] = currentWorld->isTileFree(vector2i(pos.x,pos.y+1));
			break;
		case 3:
			prog[c] = currentWorld->isTileFree(vector2i(pos.x,pos.y+1));
			break;
		}

		pc+=4;

	} else if(datType == 1)
	{
		// DAT,[type],[direction],[result]
		unsigned int b = (pc+2)%PROG_SIZE;
		unsigned int c = (pc+3)%PROG_SIZE;
		switch(prog[b]%4)
		{
		case 0:
			prog[c] = currentWorld->doesTileContainFood(vector2i(pos.x,pos.y+1));
			break;
		case 1:
			prog[c] = currentWorld->doesTileContainFood(vector2i(pos.x,pos.y+1));
			break;
		case 2:
			prog[c] = currentWorld->doesTileContainFood(vector2i(pos.x,pos.y+1));
			break;
		case 3:
			prog[c] = currentWorld->doesTileContainFood(vector2i(pos.x,pos.y+1));
			break;
		}

		pc+=4;

	}
}

void entity::EXE()
{
	// EXE,[direction]

	unsigned int a = (pc+1)%PROG_SIZE;
	switch(prog[a]%4)
	{
	case 0:
		move('n');
		break;
	case 1:
		move('e');
		break;
	case 2:
		move('s');
		break;
	case 3:
		move('w');
		break;
	}

	pc+=2;
}