#ifndef _TILE_H_
#define _TILE_H_

#include <iostream>

#include "debug.h"

#include "entity.h"

enum tileType{ Tinvalid=0, Ttile, TopenTile, TblockedTile };


class invalid
{
public:
	static bool hasFunction(tileType T) { return false; }
};

template<typename T>
struct tileTypeCheck
{
	enum { typeId = 0 };
};

#define type( type )					\
	template<>							\
	struct tileTypeCheck< type >		\
	{									\
		enum { typeId = T##type };		\
	};									

#define getTileType( T ) (tileType) tileTypeCheck< T >::typeId


#define tileFuncs( type, parent )		\
	virtual bool hasFunction(tileType T) { return (T==T##type)?true:((T##type==Ttile)?false:parent::hasFunction(T)); }	\
	virtual void* ptr() { return this; }





class tile
{
public:
	tile() {}
	virtual ~tile() {}
	tileFuncs(tile,invalid);

	template<typename T>
	void getTilePtr(T* &tPtr)
	{		
		tPtr = NULL;
		if( !hasFunction(getTileType(T)) ) return;
		tPtr = static_cast<T*>(ptr());
	}

	virtual void update()=0;
};
type(tile);







class openTile : public tile
{
public:
	openTile() { entityPtr=NULL; food=false;foodDuration=0; }
	virtual ~openTile() {}
	tileFuncs(openTile,tile);

	virtual void update();

	virtual bool isFree() { return entityPtr==NULL; }
	
	entity* getEntity() { return entityPtr; }
	void removeEntity() { entityPtr=NULL; }
	void addEntity(entity* e) { CRITICAL(!entityPtr); entityPtr=e; }
	bool addFood() { if(food) return false; food = true; return true; }
	bool containsFood() { return food; }

protected:
	entity* entityPtr;
	bool food;
	unsigned int foodDuration;
};
type(openTile);





class blockedTile : public tile
{
public:
	blockedTile() {}
	virtual ~blockedTile() {}
	tileFuncs(blockedTile,tile);

	virtual void update() {}
};
type(blockedTile);

#endif
