
#include <list>
#include <algorithm>

#include "debug.h"

#include "entity.h"
#include "world.h"
#include "tile.h"
#include "vector.h"

world *currentWorld = NULL;

world::world(	unsigned int seed,
				unsigned int length_,
				unsigned int width_,
				float openness_,
				float fertility_,
				unsigned int foodDuration_)
{
	length = length_;
	width = width_;
	openness = openness_;
	fertility = fertility_;
	foodDuration = foodDuration_;

	grid = new tile** [length];
	CRITICAL(grid);

	srand(seed);

	unsigned int openCnt = 0;
	unsigned int closedCnt = 0;
	unsigned int foodCnt = 0;

	for(unsigned int i=0;i<length;i++)
	{
		grid[i] = new tile* [width];
		CRITICAL(grid[i]);

		for(unsigned int j=0;j<width;j++)
		{
			int r = rand()%1000;
			float f = (float)r/1000.0f;
			if(frand()<openness)
			{
				openTile* openTilePtr = new openTile;
				CRITICAL(openTilePtr);

				grid[i][j] = openTilePtr;

				openCnt++;
			} else
			{
				grid[i][j] = new blockedTile;
				closedCnt++;
			}
		}
	}
	std::cout << "open tiles: " << openCnt << std::endl;
	std::cout << "closed tiles: " << closedCnt << std::endl;
	std::cout << "tiles with food: " << foodCnt << std::endl;
}

world::~world()
{
	for(unsigned int i=0;i<length;i++)
	{
		for(unsigned int j=0;j<width;j++)
		{
			delete grid[i][j];
		}

		delete [] grid[i];
	}
	delete [] grid;

	l_entitiesIt it = livingEntities.begin();
	while(it!=livingEntities.end())
	{
		livingEntities.erase(it);
		it = livingEntities.begin();
	}

	it = blackList.begin();
	while(it!=blackList.end())
	{
		blackList.erase(it);
		it = blackList.begin();
	}

	it = deadEntities.begin();
	while(it!=deadEntities.end())
	{
		deadEntities.erase(it);
		it = deadEntities.begin();
	}
}


void world::doWorld()
{
	for(unsigned int i=0;i<length;i++)
	{
		for(unsigned int j=0;j<width;j++)
		{
			grid[i][j]->update();
		}
	}
	
	if(frand()<fertility)
	{
		openTile *tPtr;
		grid[rand()%length][rand()%width]->getTilePtr(tPtr);
		if(tPtr)
		{
			tPtr->addFood();
		}
	}
}

void world::doEntities()
{
	static entity* bestEntity = NULL;

	l_entitiesIt it = livingEntities.begin();
	while(it!=livingEntities.end())
	{
		(*it)->update();
		++it;
	}

	if(blackList.size()>0)
	{
		it = blackList.begin();
		while(it!=blackList.end())
		{
			deadEntities.push_back(*it);
			l_entitiesIt lit = std::find(livingEntities.begin(),livingEntities.end(),*it);
			CRITICAL(lit!=livingEntities.end());

			livingEntities.erase(lit);
			++it;
		}

		deadEntities.sort(entity::entityCompare());


		while(deadEntities.size()>10)
		{
			deadEntities.pop_back();
		}

		if(bestEntity!=(*(deadEntities.begin())))
		{
			std::cout << "new best: " << (*(deadEntities.begin()))->age << std::endl; 
			bestEntity = (*(deadEntities.begin()));
		}
		
		blackList.clear();
	}

}



bool world::isTileFree(const vector2i& v)
{
	if(	v.x>=length || v.y>=width ) return false;

	openTile* tPtr;
	grid[v.x][v.y]->getTilePtr(tPtr);

	if(!tPtr) return false;

	return tPtr->isFree();
}

bool world::doesTileContainFood(const vector2i& v)
{
	if(	v.x>=length || v.y>=width ) return false;

	openTile* tPtr;
	grid[v.x][v.y]->getTilePtr(tPtr);

	if(!tPtr) return false;

	return tPtr->containsFood();
}

bool world::addEntity(entity& e, const vector2i& pos)
{
	if( !isTileFree(pos) ) return false;

	openTile* tPtr;
	grid[pos.x][pos.y]->getTilePtr(tPtr);
	if(!tPtr) return false;

	tPtr->addEntity(&e);

	e.pos = pos;
	livingEntities.push_back(&e);
	return true;
}

void world::removeEntity(entity& e)
{
	openTile* tPtr;
	grid[e.pos.x][e.pos.y]->getTilePtr(tPtr);
	CRITICAL(tPtr);

	tPtr->removeEntity();
}

void world::moveEntity(entity& e, const vector2i& to)
{
	if( !isTileFree(to) ) return;

	openTile* tPtr;
	grid[e.pos.x][e.pos.y]->getTilePtr(tPtr);
	CRITICAL(tPtr);

	tPtr->removeEntity();

	grid[to.x][to.y]->getTilePtr(tPtr);
	CRITICAL(tPtr);

	tPtr->addEntity(&e);

	e.pos = to;
}

void world::killEntity(entity* e)
{
	removeEntity(*e);

	blackList.push_back(e);
}
