

#ifdef WIN32
#include "Windows.h"
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "common.h"

#include "virtualMachine.h"



void instr_null(process& proc, address& pc)
{
	++pc;
}

void instr_add(process& proc, address& pc)
{
	address addr;
	int val;

	addr = proc.m_program.read(++pc);
	val = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	val += proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	proc.m_program.write(addr, val);

    ++pc;
}


void instr_sub(process& proc, address& pc)
{
	address addr;
	int val;

	addr = proc.m_program.read(++pc);
	val = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	val -= proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	proc.m_program.write(addr, val);

	++pc;
}



void instr_divide(process& proc, address& pc)
{
	address addr;
	int a, b;

	addr = proc.m_program.read(++pc);
	a = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	b = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	proc.m_program.write(addr, b?a/b:0);

	++pc;
}



void instr_multiply(process& proc, address& pc)
{
	address addr;
	int val;

	addr = proc.m_program.read(++pc);
	val = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	val *= proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	proc.m_program.write(addr, val);

	++pc;
}



void instr_copy(process& proc, address& pc)
{
	address addr;
	int val;

	addr = proc.m_program.read(++pc);
	val = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	proc.m_program.write(addr, val);

	++pc;
}



void instr_jump(process& proc, address& pc)
{
	address addr;

	addr = proc.m_program.read(++pc);
	pc = proc.m_program.read(addr);
}



void instr_jumpIfEqual(process& proc, address& pc)
{
	address addr;
	int a, b;

	addr = proc.m_program.read(++pc);
	a = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	b = proc.m_program.read(addr);
    
	addr = proc.m_program.read(++pc);

	if(a==b)
	{
		pc = proc.m_program.read(addr);
	} else
	{
		++pc;
	}
}



void instr_jumpIfNotEqual(process& proc, address& pc)
{
	address addr;
	int a, b;

	addr = proc.m_program.read(++pc);
	a = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	b = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);

	if(a!=b)
	{
		pc = proc.m_program.read(addr);
	} else
	{
		++pc;
	}
}



void instr_jumpIfGreaterThan(process& proc, address& pc)
{
	address addr;
	int a, b;

	addr = proc.m_program.read(++pc);
	a = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	b = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);

	if(a>b)
	{
		pc = proc.m_program.read(addr);
	} else
	{
		++pc;
	}
}



void instr_jumpIfLessThan(process& proc, address& pc)
{
	address addr;
	int a, b;

	addr = proc.m_program.read(++pc);
	a = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);
	b = proc.m_program.read(addr);

	addr = proc.m_program.read(++pc);

	if(a<b)
	{
		pc = proc.m_program.read(addr);
	} else
	{
		++pc;
	}
}




void instr_spawnThread(process& proc, address& pc)
{
	address addrStart, addrTid;
	uInt tid;

	addrStart = proc.m_program.read(++pc);
	addrTid = proc.m_program.read(++pc);

	tid = proc.spawnThread(addrStart);
	proc.m_program.write(addrTid, tid);

	++pc;
}



void instr_killThread(process& proc, address& pc)
{
	address addrTid;
	uInt tid;

	addrTid = proc.m_program.read(++pc);
	tid = proc.m_program.read(addrTid);

	if(proc.getNumThreads()>1)
	{
		proc.killThread(tid%proc.getMaxThreads());
	}

	++pc;
}


virtualMachine g_VM;
program g_prog;
process* g_pProc;
uInt g_seed;
uInt g_progSize;
uInt g_progDim;
uInt g_numThreads;
uInt g_windowDim[2];
uInt g_mousePos[2];
uInt g_currentPaintVal;


void idle()
{
	g_VM.step();
	glutPostRedisplay();
}


void rectangle(float x, float y, float w, float h, float lw, const float* c )
{
	static uInt displayListId = 0;

	glPushMatrix();
	glTranslatef(x,y,0.0f);
	glScalef(w,h,1.0f);

	if(c) glColor3fv(c);
	glLineWidth(lw);

	if(displayListId)
	{
		glCallList(displayListId);
	} else
	{
		displayListId = glGenLists(1);
		glNewList(displayListId, GL_COMPILE_AND_EXECUTE);

		glBegin(GL_LINE_LOOP);

		glVertex3f(0.0f,0.0f,0.0f);
		glVertex3f(1.0f,0.0f,0.0f);
		glVertex3f(1.0f,1.0f,0.0f);
		glVertex3f(0.0f,1.0f,0.0f);

		glEnd();

		glEndList();
	}

	glPopMatrix();
}


void filledRectangle(float x, float y, float w, float h, const float* c )
{
	static uInt displayListId = 0;

	glPushMatrix();
	glTranslatef(x,y,0.0f);
	glScalef(w,h,1.0f);

	if(c) glColor4fv(c);

	if(displayListId)
	{
		glCallList(displayListId);
	} else
	{
		displayListId = glGenLists(1);
		glNewList(displayListId, GL_COMPILE_AND_EXECUTE);

			glBegin(GL_POLYGON);

				glVertex2f(0.0f,0.0f);
				glVertex2f(1.0f,0.0f);
				glVertex2f(1.0f,1.0f);
				glVertex2f(0.0f,1.0f);

			glEnd();

		glEndList();
	}

	glPopMatrix();
}

template<uInt subdivs>
void squareGrid()
{
	static uInt gridDisplayListId = 0;
	if(gridDisplayListId)
	{
		glCallList(gridDisplayListId);
	} else
	{
		gridDisplayListId = glGenLists(1);
		glNewList(gridDisplayListId, GL_COMPILE_AND_EXECUTE);

		glColor3f(1.0f,1.0f,1.0f);

		glBegin(GL_LINES);

		for(uInt i=0;i<=subdivs;i++)
		{
			glVertex2f((float)i/subdivs,0);
			glVertex2f((float)i/subdivs,1);
			glVertex2f(0,(float)i/subdivs);
			glVertex2f(1,(float)i/subdivs);
		}

		glEnd();

		glEndList();
	}
}


address getSelectedAddress()
{
	float x = (float)g_mousePos[0]/g_windowDim[0];
	float y = (float)g_mousePos[1]/g_windowDim[1];

	address addr = g_progDim*x;
	addr+=g_progDim*(uInt)(g_progDim*y);
	return addr;
}

void draw()
{
	glClear(GL_COLOR_BUFFER_BIT);

	float c[4] = {0.0f,0.0f,0.0f,1.0f};
	float oneOverDim = 1.0f/g_progDim;

	{
		std::vector<int>::iterator it = g_prog.begin();
		for( uInt x=0,y=0; it!=g_prog.end(); ++it,++x)
		{
			if(x==g_progDim)
			{
				x=0;
				y++;
			}

			
			uInt data = (*it);
			c[0] = c[1] = c[2] = (float)data/g_max_uInt;
			filledRectangle(x*oneOverDim,y*oneOverDim,oneOverDim,oneOverDim,c);
		}
	}

	{
		std::vector<thread>::iterator it = g_pProc->begin();
		int x, y;

		for( uInt cnt=0; it!=g_pProc->end();++it,++cnt)
		{
			/*
			c[0] = sin((float)cnt);
			c[1] = sin((float)cnt+1);
			c[2] = sin((float)cnt+2);

			float sum = c[0] + c[1] + c[2];
			c[0] /= sum;
			c[1] /= sum;
			c[2] /= sum;
			c[3] = 1.0f;
			*/

			c[0] = 1.0f;
			c[1] = 0.0f;
			c[2] = 0.0f;
			c[3] = 1.0f;

			if(it->m_inUse)
			{
				x = it->m_PC;

				y = (x/g_progDim);
				x = (x%g_progDim);

				filledRectangle(x*oneOverDim,y*oneOverDim,oneOverDim,oneOverDim,c);
			}
		}
	}

//	glLineWidth(1.0f);
//	glColor3f(1.0f,1.0f,1.0f);
//	squareGrid<32>();

	{
		address addr = getSelectedAddress();
		int x, y;
		x = (addr%g_progDim);
		y = (addr/g_progDim);

		c[0] = 0.0f;
		c[1] = 0.0f;
		c[2] = 1.0f;
		c[3] = 1.0f;

		rectangle(x*oneOverDim,y*oneOverDim,oneOverDim,oneOverDim,1,c);

		c[3] = 0.5f;
		filledRectangle(x*oneOverDim,y*oneOverDim,oneOverDim,oneOverDim,c);
	}

	glutSwapBuffers();
}

void onKeyDown(uChar key, int x, int y)
{
	if(key==27)
	{
		exit(0);

	} else if(key=='>' || key=='.')
	{
		g_pProc->kill();
		g_prog.generate(++g_seed);
		g_pProc = g_VM.getProcess(g_VM.run(g_prog));
		ASSERT(g_pProc, "failed to create process");

	} else if(key=='<' || key==',')
	{
		g_pProc->kill();
		g_prog.generate(--g_seed);
		g_pProc = g_VM.getProcess(g_VM.run(g_prog));
		ASSERT(g_pProc, "failed to create process");
	} else if(key=='+' || key=='=')
	{
		g_currentPaintVal++;
	} else if(key=='-' || key=='_')
	{
		g_currentPaintVal--;
	}
}

void onResize(int w, int h)
{
	g_windowDim[0] = w;
	g_windowDim[1] = h;
	glViewport(0,0,w,h);
}


void onMouseEvent(int button, int state, int x, int y)
{
	if(button==0 && state==0)
	{
		address addr = getSelectedAddress();
		g_prog.write(addr,g_currentPaintVal);
	}
}

void onMouseMove(int x, int y)
{
	g_mousePos[0] = x;
	g_mousePos[1] = g_windowDim[1]-y;
}

void onMouseDrag(int x, int y)
{
	onMouseMove(x, y);
	onMouseEvent(0,0,x, y);
}

int main(int argc, char* argv[])
{

	g_numThreads = 100;
	g_VM.setMaxThreadsPerProcess(g_numThreads);

	g_VM.addInstruction(&instr_null);

	g_VM.addInstruction(&instr_add);

	g_VM.addInstruction(&instr_sub);

	g_VM.addInstruction(&instr_multiply);

	g_VM.addInstruction(&instr_divide);

	g_VM.addInstruction(&instr_copy);
	
	g_VM.addInstruction(&instr_jump);
	g_VM.addInstruction(&instr_jumpIfEqual);
	g_VM.addInstruction(&instr_jumpIfNotEqual);
	g_VM.addInstruction(&instr_jumpIfGreaterThan);
	g_VM.addInstruction(&instr_jumpIfLessThan);

	g_VM.addInstruction(&instr_spawnThread);

	g_VM.addInstruction(&instr_killThread);

	g_seed = 0;
	g_progSize = 20000;
	float tmp = g_progSize;
	g_progDim = (sqrt(tmp)+0.999f);

	g_prog.resize(g_progSize,0);
	//g_prog.generate(g_seed);


	glutInit(&argc, argv);

	g_windowDim[0] = g_windowDim[1] = 600;

	//glutInitWindowSize(g_windowDim[0],g_windowDim[1]);
	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE );
	glutCreateWindow("experiment02_01");
	glutFullScreen();

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glutIdleFunc(idle);
	glutDisplayFunc(draw);
	glutKeyboardFunc(onKeyDown);
	glutMouseFunc(onMouseEvent);
	glutMotionFunc(onMouseDrag);
	glutPassiveMotionFunc(onMouseMove);
	glutReshapeFunc(onResize);

	glClearColor(0.0f,0.0f,0.0f,0.0f);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0f,1.0f,0.0f,1.0f,-1.0f,1.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();


	g_pProc = g_VM.getProcess(g_VM.run(g_prog));
	if(!g_pProc) return 1;

	g_currentPaintVal = 0;

	glutMainLoop();
}
