
#include "instructions.h"
#include "virtualMachine.h"



void instr_null(virtualMachine& vm, process& proc, thread& t)
{
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	t = prog.read<address>(t);
}


void instr_copy(virtualMachine& vm, process& proc, thread& t)
{
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	prog.write(addr, data);

	t = next;
}


void instr_increment(virtualMachine& vm, process& proc, thread& t)
{
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);
	++data;
	prog.write(addr, data);

	t = next;
}


void instr_decrement(virtualMachine& vm, process& proc, thread& t)
{
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);
	--data;
	prog.write(addr, data);

	t = next;
}


void instr_switch(virtualMachine& vm, process& proc, thread& t)
{
	address a, b, addr;
	char data;
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	a = prog.read<address>(t);

	t+=sizeof(address);
	b = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);
		
	if(data)
	{
		t = a;
	} else
	{
		t = b;
	}
}


void instr_and(virtualMachine& vm, process& proc, thread& t)
{
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data &= prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	prog.write(addr, data);

	t = next;
}


void instr_or(virtualMachine& vm, process& proc, thread& t)
{
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data |= prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	prog.write(addr, data);

	t = next;
}



void instr_spawnThread(virtualMachine& vm, process& proc, thread& t)
{
	address next, addr;
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	if(proc.getNumThreads()<g_maxThreads)
	{
		t+=sizeof(address);
		addr = prog.read<address>(t);

		proc.spawnThread(addr);
	}

	t = next;
}



void instr_killThread(virtualMachine& vm, process& proc, thread& t)
{
	address next, addr;
	uChar data;
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	if(proc.getNumThreads()>1)
	{
		t+=sizeof(address);
		addr = prog.read<address>(t);
		data = prog.read<uChar>(addr);

		proc.killThread(data%g_maxThreads+1);
	}

	t = next;
}




