
#include "render.h"
#include "virtualMachine.h"
#include "instructions.h"


virtualMachine g_vm;
program g_prog;
uInt g_pid;

uInt g_bytesPerRow;
vector2f g_byteDimensions;


void draw()
{
	vector2f currentPos(0.0f,0.0f);
	colour4f fillColour(0.0f,0.0f,0.0f,1.0f);

	program::iterator it = g_prog.begin();
	for( ; it!=g_prog.end();++it)
	{

		fillColour.m_r = fillColour.m_g = fillColour.m_b = (float)*reinterpret_cast<uChar*>(&(*it))/255.0f;

		drawRectangle(	&currentPos,
						&g_byteDimensions,
						NULL,
						NULL,
						&fillColour						
			);

		currentPos.x+=g_byteDimensions.x;
		if(currentPos.x>=1.0f)
		{
			currentPos.x = 0.0f;
			currentPos.y+=g_byteDimensions.y;
		}
	}



	process* pProc = g_vm.getProcess(g_pid);

	ASSERT(pProc, "no process running");

	uInt numThreads = pProc->getNumThreads();
	vector2i bytePos;

	fillColour.m_r = 1.0f;
	fillColour.m_b = 0.0f;
	fillColour.m_a = 0.1f;

	address addr;
	for(uInt i=0;i<g_maxThreads;++i)
	{
		addr = pProc->getThreadAddress(i+1);
		if(addr==g_max_uInt) continue;

		addr%=g_prog.size();

		fillColour.m_g = (float)i/g_maxThreads;

		Assert(addr!=g_max_uInt, "invalid address");
		
		currentPos.x = (addr%g_bytesPerRow)*g_byteDimensions.x;
		currentPos.y = (addr/g_bytesPerRow)*g_byteDimensions.y;
		
		drawRectangle(	&currentPos,
						&g_byteDimensions,
						NULL,
						NULL,
						&fillColour						
			);
	}
}


void update()
{
	g_vm.step();
}


int main(int argc, char* argv[])
{
	initRenderingContext(argc, argv);
	setRenderCallback(draw);
	setMainLoopCallback(update);

	g_bytesPerRow = 255;
	g_byteDimensions.x = g_byteDimensions.y = 1.0f/g_bytesPerRow;

	g_prog.resize(g_bytesPerRow*g_bytesPerRow);
	g_prog.generate(0);

	g_vm.addInstruction(instr_null);
	g_vm.addInstruction(instr_copy);
	g_vm.addInstruction(instr_increment);
	g_vm.addInstruction(instr_decrement);
	g_vm.addInstruction(instr_switch);
	g_vm.addInstruction(instr_and);
	g_vm.addInstruction(instr_or);
	g_vm.addInstruction(instr_spawnThread);
	g_vm.addInstruction(instr_killThread);

	g_pid = g_vm.run(g_prog);

	beginMainLoop();

	return 0;
}

