

#include "render.h"
#include "utilities.h"

#ifdef WIN32
#include "Windows.h"
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>



vector2i g_windowDimensions;
uInt g_terrainDisplayListId;


void_func_void g_pRender;
void_func_void g_pMainLoop;


void __draw()
{
	glClear(GL_COLOR_BUFFER_BIT);
	g_pRender();
	
	glutSwapBuffers();
}

void __idle()
{
	g_pMainLoop();
	glutPostRedisplay();
}


void __reshape(int x, int y)
{
	if(x>0)
	{
		g_windowDimensions.x = x;
	} else
	{
		g_windowDimensions.x = 0;
	}
	
	if(y>0)
	{
		g_windowDimensions.y = y;
	} else
	{
		g_windowDimensions.y = 0;
	}
	
	glViewport(0, 0, g_windowDimensions.x, g_windowDimensions.y);
}


void setMainLoopCallback(void_func_void pIdle)
{
	g_pMainLoop = pIdle;
}


void setRenderCallback(void_func_void pRender)
{
	g_pRender = pRender;
}


void beginMainLoop()
{
	if(g_pMainLoop && g_pRender) glutMainLoop();
}


void getWindowDimensions( vector2i& dim )
{
	dim.x = g_windowDimensions.x;
	dim.y = g_windowDimensions.y;
}


void drawRectangle(	const vector2f* pPos,
					const vector2f* pDim,
					const float* pLineWidth,
					const colour4f* pLineColour,
					const colour4f* pFillColour)
{
	static uInt outlineDisplayListId = 0;
	static uInt fillDisplayListId = 0;

	if(pPos)
	{
		glPushMatrix();
		glLoadIdentity();
		glTranslatef(pPos->x, pPos->y, 0.0f);
		if(pDim) glScalef(pDim->x, pDim->y, 1.0f);
	}

	if(pLineColour)
	{
		glColor3fv(pLineColour->m_data);
		if(pLineWidth) glLineWidth(*pLineWidth);


		if(outlineDisplayListId)
		{
			glCallList(outlineDisplayListId);
		} else
		{
			outlineDisplayListId = glGenLists(1);
			glNewList(outlineDisplayListId, GL_COMPILE_AND_EXECUTE);

			glBegin(GL_LINE_LOOP);

			glVertex3f(0.0f,0.0f,0.0f);
			glVertex3f(1.0f,0.0f,0.0f);
			glVertex3f(1.0f,1.0f,0.0f);
			glVertex3f(0.0f,1.0f,0.0f);

			glEnd();

			glEndList();
		}
	}
	
	if(pFillColour)
	{
		glColor3fv(pFillColour->m_data);


		if(fillDisplayListId)
		{
			glCallList(fillDisplayListId);
		} else
		{
			fillDisplayListId = glGenLists(1);
			glNewList(fillDisplayListId, GL_COMPILE_AND_EXECUTE);

			glBegin(GL_POLYGON);

			glVertex3f(0.0f,0.0f,0.0f);
			glVertex3f(1.0f,0.0f,0.0f);
			glVertex3f(1.0f,1.0f,0.0f);
			glVertex3f(0.0f,1.0f,0.0f);

			glEnd();

			glEndList();
		}
	}

	if(pPos) glPopMatrix();
}


void drawGrid(	const vector2i& subdivs,
				const vector2f* pPos,
				const vector2f* pDim,
				const float* pLineWidth,
				const colour4f* pLineColour,
				const colour4f* pFillColour)
{
	
	
	if(pPos)
	{
		glLoadIdentity();
		glPushMatrix();
		glTranslatef(pPos->x, pPos->y, 0.0f);
		if(pDim) glScalef(pDim->x, pDim->y, 1.0f);
	}
	
	
	if(pLineWidth) glLineWidth(*pLineWidth);

	drawRectangle(NULL, pDim, NULL, pLineColour, pFillColour);

	glColor3fv(pLineColour->m_data);
	glBegin(GL_LINES);

	for(int i=0;i<=subdivs.x;i++)
	{
		glVertex2f((float)i/subdivs.x, 0);
		glVertex2f((float)i/subdivs.x, 1);
	}
	
	for(int i=0;i<=subdivs.y;i++)
	{
		glVertex2f(0, (float)i/subdivs.y);
		glVertex2f(1, (float)i/subdivs.y);
	}

	glEnd();
	
	
	if(pPos) glPopMatrix();
}


void initRenderingContext(int argc, char* argv[])
{
	glutInit(&argc, argv);
	
	glutInitWindowSize(800,800);
	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE );
	glutCreateWindow("experiment02_02");
	//glutFullScreen();

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glutReshapeFunc(__reshape);
	glutDisplayFunc(__draw);
	glutIdleFunc(__idle);

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}







