
#include <iostream>

#include "instructions.h"
#include "world.h"


#define process2entity(PROC) *static_cast<entity*>(&(PROC.getProgram()))
#define vm2world(VM) *static_cast<world*>(&VM)


void instr_null(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);
	
	t+=sizeof(instruction);
	t = prog.read<address>(t);
}


void instr_increment(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);
	++data;
	prog.write(addr, data);

	t = next;
}


void instr_and(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data &= prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	prog.write(addr, data);

	t = next;
}


void instr_or(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data |= prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	prog.write(addr, data);

	t = next;
}


void instr_copy(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	prog.write(addr, data);

	t = next;
}


void instr_decrement(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	address next, addr;
	char data;
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);
	--data;
	prog.write(addr, data);

	t = next;
}



void instr_add(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	program& prog = proc.getProgram();
	address next, addr;
	char val;

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);
	
	t+=sizeof(address);
	addr = prog.read<address>(t);
	val = prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	val += prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	prog.write(addr, val);

	t = next;
}


void instr_sub(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	program& prog = proc.getProgram();
	address next, addr;
	char val;

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	val = prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	val -= prog.read<char>(addr);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	prog.write(addr, val);

	t = next;
}






void instr_switch(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	address a, b, addr;
	char data;
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	a = prog.read<address>(t);

	t+=sizeof(address);
	b = prog.read<address>(t);

	t+=sizeof(address);
	addr = prog.read<address>(t);
	data = prog.read<char>(addr);

	if(data)
	{
		t = a;
	} else
	{
		t = b;
	}
}


void instr_spawnThread(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	address next, addr;
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	if(proc.getNumThreads()<g_maxThreads)
	{
		t+=sizeof(address);
		addr = prog.read<address>(t);

		proc.spawnThread(addr);
#ifdef VERBOSE
		std::cout << "pid: " << proc.getId() << "\tspawn thread: " << proc.getNumThreads() << std::endl;
#endif
	}

	
	
	t = next;
}



void instr_killThread(virtualMachine& vm, process& proc, thread& t)
{
	world& w = vm2world(vm);
	entity& ent = process2entity(proc);
	address next, addr;
	uChar data;
	program& prog = proc.getProgram();

	w.releaseEnergy(1, ent);

	t+=sizeof(instruction);
	next = prog.read<address>(t);

	if(proc.getNumThreads()>1)
	{
		t+=sizeof(address);
		addr = prog.read<address>(t);
		data = prog.read<uChar>(addr);

		proc.killThread(data%g_maxThreads+1);
#ifdef VERBOSE
		std::cout << "pid: " << proc.getId() << "\tkill thread: " << data%g_maxThreads+1 << std::endl;
#endif
	}

	
	t = next;
}


void instr_moveFoward(virtualMachine& vm, process& proc, thread& t)
{
	entity& ent = process2entity(proc);
	world& w = vm2world(vm);
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	address next = prog.read<address>(t);

	w.releaseEnergy(1, ent);
	w.moveEntity(ent);

	t = next;
}


void instr_rotate(virtualMachine& vm, process& proc, thread& t)
{
	entity& ent = process2entity(proc);
	world& w = vm2world(vm);
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	address next = prog.read<address>(t);

	t+=sizeof(address);
	address addr = prog.read<address>(t);
	char left = prog.read<char>(addr)%2;

	w.releaseEnergy(1, ent);
	if(left==1)
	{
		w.rotateEntityLeft(ent);
	} else
	{
		w.rotateEntityRight(ent);
	}

	t = next;
}



void instr_transferEnergy(virtualMachine& vm, process& proc, thread& t)
{
	entity& ent = process2entity(proc);
	world& w = vm2world(vm);
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	address next = prog.read<address>(t);

	t+=sizeof(address);
	address addr = prog.read<address>(t);
	energy e = prog.read<energy>(addr);

	w.releaseEnergy(1, ent);
	w.transferEnergy(e, ent);

	t = next;
}


void instr_absorbEnergy(virtualMachine& vm, process& proc, thread& t)
{
	entity& ent = process2entity(proc);
	world& w = vm2world(vm);
	program& prog = proc.getProgram();
	address next = prog.read<address>(t);

	w.absorbEnergy(ent);

	t = next;
}


void instr_giveMemory(virtualMachine& vm, process& proc, thread& t)
{
	entity& ent = process2entity(proc);
	world& w = vm2world(vm);
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	address next = prog.read<address>(t);

	t+=sizeof(address);
	address addr = prog.read<address>(t);
	addr = prog.read<address>(addr);

	w.releaseEnergy(1, ent);
	w.giveMemory(ent, addr);

	t = next;
}


void instr_takeMemory(virtualMachine& vm, process& proc, thread& t)
{
	entity& ent = process2entity(proc);
	world& w = vm2world(vm);
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	address next = prog.read<address>(t);

	t+=sizeof(address);
	address addr = prog.read<address>(t);
	addr = prog.read<address>(addr);

	w.releaseEnergy(1, ent);
	w.takeMemory(ent, addr);

	t = next;
}


void instr_convertEnergyToMemory(virtualMachine& vm, process& proc, thread& t)
{
	entity& ent = process2entity(proc);
	world& w = vm2world(vm);
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	address next = prog.read<address>(t);

	t+=sizeof(address);
	address addr = prog.read<address>(t);
	energy e = prog.read<energy>(addr);

	w.releaseEnergy(1, ent);
	w.convertEnergyToMemory(ent, e);

	t = next;
}


void instr_convertMemoryToEnergy(virtualMachine& vm, process& proc, thread& t)
{
	entity& ent = process2entity(proc);
	world& w = vm2world(vm);
	program& prog = proc.getProgram();

	t+=sizeof(instruction);
	address next = prog.read<address>(t);

	t+=sizeof(address);
	address addr = prog.read<address>(t);
	addr = prog.read<address>(addr);

	w.releaseEnergy(1, ent);
	w.convertMemoryToEnergy(ent, addr);

	t = next;
}


