

#ifndef __instructions_h__
#define __instructions_h__



#include "common.h"
#include "virtualMachine.h"
#include "program.h"


#include "instructions.h"
#include "world.h"

/// convert a process to an entity
#define process2entity(PROC) *static_cast<entity*>(&(PROC.getProgram()))

/// convert from a virtual machine to a world
#define vm2world(VM) *static_cast<world*>(&VM)


/// maximum number of threads per process
/**
 *	]0, 255]
 */
static const uChar g_maxThreads = 255;


/// do nothing
void instr_null(virtualMachine& vm, process& proc, thread& t);
/// logically AND two values and store the result
void instr_and(virtualMachine& vm, process& proc, thread& t);
/// logically OR two values and store the result
void instr_or(virtualMachine& vm, process& proc, thread& t);
/// copy data between addresses
void instr_copy(virtualMachine& vm, process& proc, thread& t);
/// increment a value
void instr_increment(virtualMachine& vm, process& proc, thread& t);
/// decrement a value
void instr_decrement(virtualMachine& vm, process& proc, thread& t);
/// sum two values and store the result
void instr_add(virtualMachine& vm, process& proc, thread& t);
/// subtract two values and store the result
void instr_sub(virtualMachine& vm, process& proc, thread& t);
/// test a value to determine the next instruction to execute
void instr_switch(virtualMachine& vm, process& proc, thread& t);
/// spawn a new thread
void instr_spawnThread(virtualMachine& vm, process& proc, thread& t);
/// kill a thread
void instr_killThread(virtualMachine& vm, process& proc, thread& t);
/// move foward
void instr_moveFoward(virtualMachine& vm, process& proc, thread& t);
/// rotate left or right
void instr_rotate(virtualMachine& vm, process& proc, thread& t);
/// transfer energy to another entity
void instr_transferEnergy(virtualMachine& vm, process& proc, thread& t);
/// absorb energy from the world
void instr_absorbEnergy(virtualMachine& vm, process& proc, thread& t);
/// transfer memory to the entity/tile infront
void instr_giveMemory(virtualMachine& vm, process& proc, thread& t);
/// take memory from the entity infront
void instr_takeMemory(virtualMachine& vm, process& proc, thread& t);
/// convert energy to memory
void instr_convertEnergyToMemory(virtualMachine& vm, process& proc, thread& t);
/// convert memory to energy
void instr_convertMemoryToEnergy(virtualMachine& vm, process& proc, thread& t);

#endif

