
#include "common.h"
#include "world.h"
#include "instructions.h"
#include "render.h"

/// a random seed
uInt g_seed = 0;

/// pointer to the world
world* g_pWorld = NULL;


/// creates a world
void initWorld()
{
	// create a new world
	g_pWorld = new world;

	// set some initial parameters
	g_pWorld->setEnergyPerUnitMemory(2);
	g_pWorld->setEnergyPerAbsorb(5);
	g_pWorld->setMaxEnergyPerTile(100000);
	g_pWorld->setMaxCostPerTile(1.0f);
	g_pWorld->setMaxEnergyPerEntity(20000);
	g_pWorld->setMaxMemoryPerEntity(10000);

	// create the instruction set

	// memory manipulation
	g_pWorld->addInstruction(&instr_null);
	g_pWorld->addInstruction(&instr_copy);
	g_pWorld->addInstruction(&instr_increment);
	g_pWorld->addInstruction(&instr_decrement);
	g_pWorld->addInstruction(&instr_add);
	g_pWorld->addInstruction(&instr_sub);

	// control
	g_pWorld->addInstruction(&instr_switch);

	// threading
	g_pWorld->addInstruction(&instr_spawnThread);
	g_pWorld->addInstruction(&instr_killThread);

	// movement
	g_pWorld->addInstruction(&instr_moveFoward);
	g_pWorld->addInstruction(&instr_rotate);

	// energy transfer
	g_pWorld->addInstruction(&instr_transferEnergy);
	g_pWorld->addInstruction(&instr_absorbEnergy);

	// memory transfer
	g_pWorld->addInstruction(&instr_giveMemory);
	g_pWorld->addInstruction(&instr_takeMemory);

	// conversion
	g_pWorld->addInstruction(&instr_convertEnergyToMemory);
	g_pWorld->addInstruction(&instr_convertMemoryToEnergy);

	// initialise the world
	g_pWorld->init();

	float prob = (float)rand()/RAND_MAX/2+0.5;

	// populate the world
	for(int i=0;i<100;++i)
		for(int j=0;j<100;++j)
			if((float)rand()/RAND_MAX>prob)
				g_pWorld->spawnEntity(1000, ((float)rand()/RAND_MAX)*255*255, position(i,j), g_north);

	++g_seed;
}


void draw()
{
	if(g_pWorld) g_pWorld->draw();
}


void idle()
{
	// if the world is dead or doesn't exist
	if(!g_pWorld || g_pWorld->isDead())
	{
		// delete the world
		if(g_pWorld) delete g_pWorld;

		// and create a new one
		initWorld();
	}

	// update the world
	g_pWorld->step();
}


int main(int argc, char* argv[])
{

	if(argc>1) g_seed = atoi(argv[1]);

	srand(g_seed);

	// initialise the renderer
	initRenderingContext(argc, argv);

	// set the callbacks
	setRenderCallback(draw);
	setMainLoopCallback(idle);
	
	// run the app
	beginMainLoop();
}
