
#include "world.h"
#include "render.h"



#define process2entity(PROC) *static_cast<entity*>(&(PROC.getProgram()))
#define vm2world(VM) *static_cast<world*>(&VM)


entity::entity(	uInt seed,
				const energy& e,
				const memory& m,
				const position& p,
				const direction& d	)
				: m_energy(e)
				, m_position(p)
				, m_direction(d)
{
	resize(m);
	generate(seed);
}



void entity::draw()
{
}



tile::tile(float cost, energy e)	: m_cost(cost)
									, m_energy(e)
									, m_pEntity(NULL)
{	
}




world::world()	: m_size(100,100)
				, m_absorbAmount(4)
				, m_energyPerUnitMemory(1)
				, m_maxEnergyPerTile(g_max_int)
				, m_maxCostPerTile(1.0f)
				, m_maxEnergyPerEntity(g_max_int)
				, m_maxMemoryPerEntity(10000)
{
}


world::~world()
{
	// delete all entites
	vv_tiles::iterator it = m_tiles.begin();
	for( ; it!=m_tiles.end(); ++it)
	{
		v_tiles::iterator iit = it->begin();
		for( ; iit!=it->end();++iit)
		{
			if(iit->m_pEntity) delete iit->m_pEntity;
		}
	}
}


void world::init()
{
	// randomise the tile data
	m_tiles.resize(m_size.x);

	vv_tiles::iterator it = m_tiles.begin();
	v_tiles::iterator tit;
	for( ;it!= m_tiles.end(); ++it)
	{
		it->resize(m_size.y);

		tit = it->begin();
		for( ; tit!=it->end();++tit)
		{
			tit->m_cost = (float)rand()/RAND_MAX*m_maxCostPerTile;
			tit->m_energy = (float)rand()/RAND_MAX*m_maxEnergyPerTile;
		}
	}
}


instruction world::addInstruction(instructionPtr instrPtr)
{ 
	return virtualMachine::addInstruction(instrPtr); 
}


bool world::replaceInstruction(instruction instr, instructionPtr instrPtr)
{ 
	return virtualMachine::replaceInstruction(instr, instrPtr);
}


uInt world::getNumInstructions() 
{ 
	return static_cast<uInt>(size());
}


void world::step()
{
	virtualMachine::step();
}


void world::kill(entityId entId)
{
	virtualMachine::kill(entId);
}


entity* world::spawnEntity(	energy e,
							memory m,
							const position& p,
							const direction& d	)
{
	// make sure the limits are not exceeded
	if(e>m_maxEnergyPerEntity) e = m_maxEnergyPerEntity;
	if(m>m_maxMemoryPerEntity) m = m_maxMemoryPerEntity;

	static uInt seed = 0;

	// check the position is valid and not inhabited
	if(p.x>=0 && p.x<m_size.x && p.y>=0 && p.y<m_size.y && !getEntityAtPos(p))
	{
		// create an entity
		entity* pEntity = new entity(seed++, e, m, p, d);
		if(pEntity)
		{
			// start executing its program
			pEntity->m_pid = run(*pEntity);

			// error check
			if(!pEntity->m_pid)
			{
				delete pEntity;
				pEntity = NULL;
			}

			// add the entity to the tile
			m_tiles[p.x][p.y].m_pEntity = pEntity;

			return(pEntity);
		}
	}

	return NULL;
}



entity* world::getEntity(entityId entId)
{
	process* proc = getProcess(entId);
	if(proc)
	{
		return static_cast<entity*>(&(proc->getProgram()));
	}
	return NULL;
}


entity* world::getEntityAtPos(const position& p)
{

	if(p.x>=0 && p.x<m_size.x && p.y>=0 && p.y<m_size.y)
	{
		return m_tiles[p.x][p.y].m_pEntity;
	}
	
	return NULL;
}


void world::moveEntity(entity& ent)
{
	// get the position of the tile to move to
	position newPos = getPosInfrontOfEntity(ent, 1);
	// get the entity's current tile
	tile* currentTile = getTileAtPos(ent.m_position);
	// get the tile to move to
	tile* newTile = getTileAtPos(newPos);

	ASSERT(currentTile, "entity must inhabit tile");
	
	// check if the tile exists and is emoty
	if(newTile && !getEntityAtPos(newPos))
	{
		// remove the entity from the current tile
		currentTile->m_pEntity = NULL;
		// add it to the new tile
		newTile->m_pEntity = &ent;
		// update the entity's position
		ent.m_position = newPos;

		// calculate the energy cost of moving
		// the entity
		energy cost = (energy)(
			ent.size()*(
				currentTile->m_cost-newTile->m_cost
			)
		);

		// transfer that energy from the entity
		// to the tile it just vacated
		transferEnergy(cost, *currentTile, ent);
	}

	ASSERT(ent.m_position.x>=0, "invalid coordinate");
	ASSERT(ent.m_position.x<m_size.x, "invalid coordinate");
	ASSERT(ent.m_position.y>=0, "invalid coordinate");
	ASSERT(ent.m_position.y<m_size.y, "invalid coordinate");
}


void world::rotateEntityLeft(entity& ent)
{
	int tmp = ent.m_direction.x;
	ent.m_direction.x = -ent.m_direction.y;
	ent.m_direction.y = tmp;
}

void world::rotateEntityRight(entity& ent)
{
	int tmp = ent.m_direction.x;
	ent.m_direction.x = ent.m_direction.y;
	ent.m_direction.y = -tmp;
}


energy world::transferEnergy(	energy amount,
								energy& a,
								energy& b	)
{
	// fast out
	if(amount==0) return 0;

	// make sure the amount to transfer 
	// does not exceed the amount availavle
	if(amount>a)
	{
		amount = a;

	} else if(-amount>b)
	{
		amount = -b;
	}

	// transfer the energy
	a -= amount;
	b += amount;

	// return the amount transfered
	return amount;
}

energy world::transferEnergy(	energy amount, 
								tile& t, 
								entity& ent	)
{
	// fast out
	if(amount==0) return 0;

	// make sure the maximums are not going to be
	// exceeded
	if(amount>0 && (amount+ent.m_energy)>m_maxEnergyPerEntity)
	{
		amount = m_maxEnergyPerEntity-ent.m_energy;

	} else if((amount-t.m_energy)<-m_maxEnergyPerTile)
	{
		amount = -(m_maxEnergyPerTile-t.m_energy);
	}

	// transfer the energy and return the amount
	// transfered
	return transferEnergy(amount, t.m_energy, ent.m_energy);
}


energy world::transferEnergy(	energy amount, 
								entity& entA, 
								entity& entB	)
{
	// fast out
	if(amount==0) return 0;

	// make sure the maximums are not going to be
	// exceeded
	if(amount>0 && (amount+entB.m_energy)>m_maxEnergyPerEntity)
	{
		amount = m_maxEnergyPerEntity-entB.m_energy;

	} else if((amount-entA.m_energy)<-m_maxEnergyPerEntity)
	{
		amount = -(m_maxEnergyPerEntity-entA.m_energy);
	}

	// transfer the energy and return the amount
	// transfered
	return transferEnergy(amount, entA.m_energy, entB.m_energy);
}


energy world::transferEnergy(energy amount, entity& ent)
{
	// get the tile infront of the entity
	position infront = getPosInfrontOfEntity(ent, 1);
	tile* tileInfront = getTileAtPos(infront);
	
	// check if the tile exists
	if(tileInfront && tileInfront->m_pEntity)
	{
		// transfer the energy and return the amount
		// transfered
		return transferEnergy(amount, ent, *(tileInfront->m_pEntity));
	}

	return 0;
}


void world::absorbEnergy(entity& ent)
{
	tile* currentTile = getTileAtPos(ent.m_position);
	ASSERT(currentTile, "entity must inhabit tile");

	// calculate the tile energy ratio
	energy halfMax = m_maxEnergyPerTile/2;
	float tileEnergyRatio = (float)(currentTile->m_energy-halfMax)/halfMax;

	// calculate the energy to be transfered
	energy amount = m_absorbAmount*tileEnergyRatio;

	// transfer the energy
	transferEnergy(amount, *currentTile, ent);
}


energy world::releaseEnergy(energy amount, entity& ent)
{
	tile* currentTile = getTileAtPos(ent.m_position);
	ASSERT(currentTile, "entity must inhabit tile");

	// make sure entity's don't try anything too clever
	if(amount<=0) return 0;

	// transfer the energy and return the amount
	// released
	return -transferEnergy(-amount, *currentTile, ent);
}


void world::transferMemory(address addr, program& a, program& b)
{
	uInt sizeA = static_cast<uInt>(a.size());
	uInt sizeB = static_cast<uInt>(b.size());

	// make sure the address is in range
	addr%=sizeA;

	// up-size the destination
	b.resize(sizeB+sizeA-addr);

	// copy the data
	for(address from=addr, to=sizeB; from<sizeA; ++from, ++to)
	{
		b[to] = a[from];
	}

	// down-size the source
	a.resize(addr);
}


void world::transferMemory(address addr, entity& entA, entity& entB)
{
	int sizeA = static_cast<int>(entA.size());
	int sizeB = static_cast<int>(entB.size());

	// fast out
	if(sizeA<=0) return;


	// make sure the address is in range
	addr%=sizeA;

	memory amountToTransfer = sizeA-addr;

	// make sure maximum will not be exceeded
	if(amountToTransfer+sizeB>m_maxMemoryPerEntity)
	{
		amountToTransfer = m_maxMemoryPerEntity-sizeB;
	}

	// calculate the relative height of the tiles the
	// entites inhabit
	tile* tileA = getTileAtPos(entA.m_position);
	tile* tileB = getTileAtPos(entB.m_position);

	ASSERT(tileA, "entity must inhabit tile");
	ASSERT(tileB, "entity must inhabit tile");

	float deltaCost = tileB->m_cost-tileA->m_cost;

	// calculate the transfer cost
	energy costToTransfer = (energy)(amountToTransfer*deltaCost);

	if(costToTransfer>0)
	{
		// if the cost is positive then the entity doing
		// the transfer looses this energy to its tile
		costToTransfer = releaseEnergy(costToTransfer, entA);

	} else if(costToTransfer<0)
	{
		// if the cost is negative then the entity gaining
		// the memory gains the energy
		costToTransfer = transferEnergy(costToTransfer, entB, entA);
	} else
	{
		// otherwise there is no cost
		deltaCost = 1.0f;
	}

	// re-evaluate the cost to transfer the memory from
	// the energy (to avoid rounding errors)
	amountToTransfer = (memory)(costToTransfer/deltaCost);

	// transfer the memory
	entB.resize(sizeB+amountToTransfer);

	for(int from=sizeA-amountToTransfer, to=sizeB; from<sizeA; ++from, ++to)
	{
		entB[to] = entA[from];
	}

	entA.resize(sizeA-amountToTransfer);
}


void world::giveMemory(entity& ent, address addr)
{
	position infront = getPosInfrontOfEntity(ent, 1);
	tile* tileInfront = getTileAtPos(infront);

	if(tileInfront && tileInfront->m_pEntity)
	{
		// if the tile infront is inhabited then
		// transfer the memory to the entity in that tile
		transferMemory(addr, ent, *(tileInfront->m_pEntity));

	} else
	{
		// otherwise create a new entity in the tile
		entity* newEnt = spawnEntity(0, 0, infront, ent.m_direction);
		if(newEnt)
		{
			// and give it the memory
			transferMemory(addr, ent, *newEnt);
		}
	}
}


void world::takeMemory(entity& ent, address addr)
{
	position infront = getPosInfrontOfEntity(ent, 1);
	tile* tileInfront = getTileAtPos(infront);

	if(tileInfront && tileInfront->m_pEntity)
	{
		// if there is an entity, then grab the memory
		transferMemory(addr, *(tileInfront->m_pEntity), ent);
	}
}

	
void world::convertEnergyToMemory(entity& ent, energy amount)
{
	// fast out
	if(ent.m_energy<=0) return;

	// no funny stuff
	if(amount<0) amount = -amount;

	memory current = static_cast<uInt>(ent.size());

	// make sure the amount of energy is in range
	amount%=ent.m_energy;

	// calculate the energy as memory
	memory toAdd = amount/m_energyPerUnitMemory;

	// fast out
	if(toAdd==0) return;
	
	// make sure maximum will not be exceeded
	if(current + toAdd > m_maxMemoryPerEntity)
	{
		toAdd = m_maxMemoryPerEntity-current;
	}

	// recalculate the amount from the memory
	// to avoid rounding errors
	amount = toAdd*m_energyPerUnitMemory;

	// remove the energy
	ent.m_energy-=amount;
	
	// add the memory and fill
	// it with null instructions
	ent.resize(current+toAdd, 0);
}


void world::convertMemoryToEnergy(entity& ent, address addr)
{
	// fast out
	if(ent.size()==0) return;

	// make sure the address is in range
	memory current = static_cast<uInt>(ent.size());
	addr%=current;
	
	// calculate the amount of memory to convert
	// and the amount of energy
	memory toConvert = current - addr;
	energy toAdd = toConvert*m_energyPerUnitMemory;

	// make sure the maximum will not be exceeded
	if(toAdd+ent.m_energy+m_energyPerUnitMemory>m_maxEnergyPerEntity)
	{
		toAdd = m_maxEnergyPerEntity-ent.m_energy-m_energyPerUnitMemory;
	}

	// add the energy
	ent.m_energy += toAdd;

	// recalculate the memory from the energy
	toConvert = toAdd/m_energyPerUnitMemory;
	
	// remove the memory
	ent.resize(current-toConvert);
	
	// give any the energy not converted
	// back to the entity
	ent.m_energy+=toAdd%m_energyPerUnitMemory;
}


tile* world::getTileAtPos(const position& p)
{
	if(p.x>=0 && p.x<m_size.x && p.y>=0 && p.y<m_size.y)
	{
		return &(m_tiles[p.x][p.y]);
	}	
	return NULL;
}


position world::getPosInfrontOfEntity(entity& ent, uInt d)
{
	position pos;
	pos.x = ent.m_position.x+d*ent.m_direction.x;
	pos.y = ent.m_position.y+d*ent.m_direction.y;
	return pos;
}





void world::draw()
{
	float tileSize = 1.0f/m_size.x;
	
	vector2f pos, dim;

	dim.x = dim.y = tileSize;
	float lineWidth = 1.0f;

	colour4f lineColour(1.0f,1.0f,1.0f,1.0f);
	colour4f tileColour(0.0f,1.0f,0.0f,1.0f);
	colour4f entityColour(1.0f,0.0f,0.0f,1.0f);

	vv_tiles::iterator xIt = m_tiles.begin();
	v_tiles::iterator yIt;
	for( uInt x=0; xIt!=m_tiles.end(); ++xIt, ++x)
	{
		yIt = xIt->begin();
		for( uInt y=0; yIt!=xIt->end(); ++yIt, ++y)
		{
			pos.x = x*tileSize;
			pos.y = y*tileSize;

			if(yIt->m_pEntity==NULL)
			{
				tileColour.m_g = yIt->m_cost/m_maxCostPerTile;
				drawRectangle(&pos, &dim, NULL, NULL, &tileColour);
			} else
			{
				entityColour.m_b = (float)yIt->m_pEntity->m_energy/1000.0f;
				drawRectangle(&pos, &dim, NULL, NULL, &entityColour);
			}
		}
	}

	drawGrid(m_size, NULL, NULL, &lineWidth, &lineColour, NULL);
}



bool world::canBeExecuted(process& proc)
{
	if(proc.isTerminated()) return false;

	entity& ent = process2entity(proc);
	return ent.m_energy > 0;
}




void world::kill(process& proc)
{	
	entity* pEnt = &process2entity(proc);

	
	
	tile* t = getTileAtPos(pEnt->m_position);
	ASSERT(t->m_pEntity==pEnt, "entity not in specified tile");
	t->m_pEntity = NULL;
	t->m_energy += pEnt->m_energy;

#ifdef VERBOSE	
	std::cout << "pid: " << proc.getId() << "\tkill\tE:" << pEnt->m_energy << "\tM:" << pEnt->size() << std::endl;
#endif
	
	delete pEnt;

	
	virtualMachine::kill(proc);
	
}

