

#ifndef __world_h__
#define __world_h__


#include "virtualMachine.h"
#include "utilities.h"




/// holds the data for an entity
class entity
	: public program
{
	friend class world;

public:
	entity(	uInt seed,
			const energy& e,
			const memory& m,
			const position& p,
			const direction& d	);

	~entity() {}

	energy getEnergy() { return m_energy; }
	virtual void draw();

protected:

	/// the entity's energy pool
	energy m_energy;
	/// the direction the entity is facing
	direction m_direction;
	/// the current position of the entity
	position m_position;
	/// the entity's process in the virtual
	/// machine
	processId m_pid;

private:
};


/// holds the data for a tile
class tile
{
	friend class world;

public:
	tile()	: m_pEntity(NULL)
			, m_cost(0.0f)
			, m_energy(0) {}

	tile(float cost, energy e);
	~tile() {}

	float getCost() { return m_cost; }

protected:
	/// the entity currently in the tile
	entity* m_pEntity;
	/// the height of the tile
	float m_cost;
	/// the tile's energy pool
	energy m_energy;

private:
};



class world
	: public virtualMachine
{
public:
	world();
	~world();

	void init();

	instruction addInstruction(instructionPtr instrPtr);
	bool replaceInstruction(instruction instr, instructionPtr instrPtr);
	uInt getNumInstructions();

	void setMaxThreadsPerProcess(uInt n);
	uInt getMaxThreadsPerProcess();

	void setEnergyPerUnitMemory(energy e) { m_energyPerUnitMemory = e; }
	energy getEnergyPerUnitMemory() { return m_energyPerUnitMemory; }

	void setEnergyPerAbsorb(energy e) { m_absorbAmount = e; }						
	energy getEnergyPerAbsorb() { return m_absorbAmount; }

	void setMaxEnergyPerTile(energy e) { m_maxEnergyPerTile = e; }
	void setMaxCostPerTile(float e) { m_maxCostPerTile = e; }
	void setMaxEnergyPerEntity(energy e) { m_maxEnergyPerEntity = e; }
	void setMaxMemoryPerEntity(memory m) { m_maxMemoryPerEntity = m; }

	void step();

	entity* spawnEntity(energy e, memory m, const position& p, const direction& d	);
	void kill(entityId entId);
	entity* getEntity(entityId entId);
	entity* getEntityAtPos(const position& p);
	
	void moveEntity(entity& ent);
	void rotateEntityLeft(entity& ent);
	void rotateEntityRight(entity& ent);

	energy transferEnergy(energy amount, energy& a, energy& b);
	energy transferEnergy(energy amount, tile& t, entity& ent);
	energy transferEnergy(energy amount, entity& entA, entity& entB);
	energy transferEnergy(energy amount, entity& ent);

	void absorbEnergy(entity& ent);
	energy releaseEnergy(energy amount, entity& ent);

	void transferMemory(address addr, program& a, program& b);
	void transferMemory(address addr, entity& entA, entity& entB);
	void giveMemory(entity& ent, address addr);
	void takeMemory(entity& ent, address addr);

	void convertEnergyToMemory(entity& ent, energy e);
	void convertMemoryToEnergy(entity& ent, address addr);
	
	tile* getTileAtPos(const position& p);
	position getPosInfrontOfEntity(entity& ent, uInt d);

	virtual void draw();

	bool isDead() { return !activeProcesses(); }

protected:
	virtual bool canBeExecuted(process& proc);
	virtual void kill(process& proc);

	typedef std::vector<tile> v_tiles;
	typedef std::vector<v_tiles> vv_tiles;
	vv_tiles m_tiles;
	vector2i m_size;

	energy m_absorbAmount;
	energy m_energyPerUnitMemory;

	energy m_maxEnergyPerTile;
	float m_maxCostPerTile;

	energy m_maxEnergyPerEntity;
	energy m_maxMemoryPerEntity;

private:
};



#endif

