/********************************************************************************
 *
 *  urbanVictory engine
 *  Copyright (C) 2003, 2004 Adam Cubitt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *										
 ********************************************************************************/

/**
 *	\file	rectangle.h
 *	\author	Adam Cubitt
 *	\brief	defines a basic axis-aligned rectangle class
 */


#ifndef __RECTANGLE__
#define __RECTANGLE__


#include "transform.h"


/// XY-axis aligned rectangle
class AArectangle
	: public transform<noRotation>
{
public:
	AArectangle() : m_width(1.0f), m_height(1.0f) {}
	virtual ~AArectangle() {}

	/// set the dimensions
	void setDim(v2f dim) { m_width = dim.x; m_height = dim.y; }
	/// set the position
	void setPos(v2f pos) { m_translation.x = pos.x, m_translation.y = pos.y; }
	/// get the width
	float getWidth() const { return m_width; }
	/// get the height
	float getHeight() const { return m_height; }
	/// query the dimensions
	v2f getDim() const { return v2f(m_width, m_height); }
	/// query the position
	v3f getPos() const { return m_translation; }

protected:
	/// the rectangles width
	float m_width;
	/// the rectangles height
	float m_height;

private:
};


#endif
