/********************************************************************************
 *
 *  urbanVictory engine
 *  Copyright (C) 2003, 2004 Adam Cubitt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *										
 ********************************************************************************/

/**
 *	\file	transform.h
 *	\author	Adam Cubitt
 *	\brief	defines transform classes
 */




#ifndef __TRANSFORM__
#define __TRANSFORM__


#include "vector.h"
#include "matrix.h"
#include "quaternion.h"


class quaternionRotation;

/// transform
/**
 *	template arguments:
 *	\arg orientationPolicy : determines how the transform handles
 *								rotation data (default:
 *								quaternionRotaton)
 */
template <
	class orientationPolicy = quaternionRotation
>
class transform : public orientationPolicy
{
public:
	transform() {}
	virtual ~transform() {}

	void toMatrix(matrix& M) const
	{
		// rotate
		orientationPolicy::toMatrix(M);
		// translate
		M[12] = m_translation.x;
		M[13] = m_translation.y;
		M[14] = m_translation.z;
	}

	matrix getAsMatrix() const
	{
		matrix M;
		toMatrix(M);
		return M;
	}

	void queryTranslation(v3f& T) const
	{
		T = m_translation;
	}

	v3f getTranslation() const
	{
		return m_translation;
	}

	void setTranslation(const v3f& t)
	{
		m_translation = t;
	}

	void setFrom(const matrix& M)
	{
		// rotation
		orientationPolicy::setFrom(M);
		// translation
		m_translation = v3f(M[12],M[13],M[14]);
	}

	v3f m_translation;
protected:
private:
};





/// orientataion is stored as a quaternion
class quaternionRotation
{
public:
	void toMatrix(matrix& M) const;
	quaternion getOrientation() const;
	void queryOrientation(quaternion& q) const;
	void queryOrientation(v3f& e) const;
	void setFrom(const matrix& M);
	void setOrientation(const quaternion& q);
	void setOrientation(const v3f& q);

	quaternion m_orientation;
protected:
private:
};






/// orientation is stored as an euler rotataion
/**
 *	\warning converting from euler to other methods of encoding
 *				rotation can be unreliable
 */
class eulerRotation
{
public:
	void toMatrix(matrix& M) const;
	quaternion getOrientation() const;
	void queryOrientation(quaternion& q) const;
	void queryOrientation(v3f& e) const;
	void setFrom(const matrix& M);
	void setOrientation(const quaternion& q);
	void setOrientation(const v3f& q);


	v3f m_orientation;
protected:
private:
};


class noRotation
{
public:
	void toMatrix(matrix& M) const;
	quaternion getOrientation() const;
	void queryOrientation(quaternion& q) const;
	void queryOrientation(v3f& e) const;
	void setFrom(const matrix& M);
	void setOrientation(const quaternion& q);
	void setOrientation(const v3f& q);


protected:
private:
};
#endif

