

#ifndef __virtualMachine_h__
#define __virtualMachine_h__


#include "common.h"
#include "id.h"
#include "program.h"



typedef uInt instruction;
typedef uInt threadId;
typedef uInt processId;




class thread
{
public:
	thread()	: m_PC(0)
				, m_inUse(false) {}

	thread(address PC)	: m_PC(PC)
						, m_inUse(true) {}

	~thread() {}

	void spawn(const address& PC) { m_PC = PC; m_inUse = true; }
	void kill() { m_inUse = false; }

	operator address() { return m_PC; }

	address m_PC;
	bool m_inUse;

protected:
private:
};


class process
	: public std::vector<thread>
{
public:
	typedef void(*instructionPtr)(process&, address&);
	typedef std::vector<instructionPtr> instructionSet;

	process(uInt maxThreads, processId pid, program& prog);
	~process() { clear(); }

	void kill() { m_terminated = true; }
	processId queryId() { return m_pid; }
	threadId spawnThread(address addr);
	void killThread(threadId tid);
	uInt getNumThreads() { return m_threadCnt; }
	uInt getMaxThreads() { return static_cast<uInt>(size()); }
	void step(instructionSet& instrSet);
	bool isTerminated() { return m_terminated || m_program.size()==0; }

	program& m_program;

protected:
	processId m_pid;
	uInt m_threadCnt;
	bool m_terminated;

private:
};


typedef void(*instructionPtr)(process&, address&);
typedef std::vector<instructionPtr> instructionSet;




class virtualMachine
	: protected instructionSet
{
public:



	virtualMachine()	: m_maxThreadsPerProcess(1)
						, m_usedCycles(1.0f)
						, m_idleCnt(0)
						, m_numProcesses(0) {}

	~virtualMachine();

	void clean();

	instruction addInstruction(instructionPtr instrPtr);
	bool replaceInstruction(instruction instr, instructionPtr instrPtr);

	processId run(program& prog);
	void kill(processId pid);
	process* getProcess(processId pid);
	uInt getNumProcesses() { return m_numProcesses; }
	void step();

	void setMaxThreadsPerProcess(uInt numThreads);
	threadId spawnThread(processId pid, address addr);
	void killThread(threadId tid);
	void setNumUsedCycles(float num);

	virtual void update(float deltaT);

protected:
	std::vector<process*> m_pProcesses;
	uInt m_maxThreadsPerProcess;
	idGen<recycleReleasedIds, noWrapAround> m_processIdGen;
	float m_usedCycles;
	uInt m_idleCnt;
	uInt m_numProcesses;

private:
};



#endif

