
#include "process.h"





process::process(program& prog, processId pid)
	: m_pProgram(&prog)
	, m_id(pid)
	, m_threadCnt(1)
{
	push_back(new thread(0));
}


process::~process()
{
	clean();
}


void process::clean() 
{ 
	// delete all threads
	iterator tit = begin();
	for( ; tit!=end();++tit)
	{
		if(*tit) delete (*tit);
	}

	resize(0);

	// clear all data
	m_threadCnt = m_id = 0; 
	m_pProgram = NULL;
}


program& process::getProgram() 
{ 
	ASSERT(m_pProgram, "cannot manipulate terminated process");
	return *m_pProgram; 
}


processId process::getId()
{
	return m_id; 
}


threadId process::spawnThread(address addr)
{
	// check if there is an unused thread
	if(m_threadCnt<size())
	{
		// find the unused thread
		iterator it = begin();
		for(uInt tid=0; it!=end(); ++tid, ++it)
		{
			if((*it)->isTerminated())
			{
				// respawn the thread
				(*it)->spawn(addr);

				m_threadCnt++;
				return tid+1;
			}
		}
	}

	// create a new thread
	push_back(new thread(addr));
	m_threadCnt++;

	return static_cast<uInt>(size());
}


void process::killThread(threadId tid)
{
	// verify that the thread exists
	if(tid>0 && tid<=size() && !(*this)[tid-1]->m_terminated)
	{
		// terminate the thread
		(*this)[tid-1]->m_terminated=true;
		m_threadCnt--;
	}
}


uInt process::getNumThreads() 
{
	return m_threadCnt;
}	


address process::getThreadAddress(threadId tid)
{
	// verify that the thread exists
	if(tid>0 && tid<=size() && !(*this)[tid-1]->m_terminated)
	{
		// return its address
		thread &t = *((*this)[tid-1]);
		if(!t.isTerminated()) return t;
	}

	return g_max_uInt;
}



bool process::isTerminated() 
{
	return		m_threadCnt==0 
			||	m_pProgram==NULL 
			||	m_pProgram->size()==0;
}


process::process()
	: m_pProgram(NULL)
	, m_id(0) 
{
}





