

#ifndef __process_h__
#define __process_h__

#include "common.h"
#include "id.h"
#include "program.h"
#include "thread.h"



/// holds the data for a process
/*
*	a process is a collection of one or more threads
*/
class process
	: protected std::vector<thread*>
{
	friend class virtualMachine;
	friend class std::vector<thread*>;

public:
	typedef void(*instructionPtr)(virtualMachine&, process&, thread&);
	typedef std::vector<instructionPtr> instructionSet;

	/// initialise active process
	/*
	*	creates a thread at address 0 of the program
	*/
	process(program& prog, uInt pid);
	~process();

	/// terminates and cleans up the thread
	void clean();

	/// get the program being run
	program& getProgram();

	/// get the id of the process
	processId getId();

	/// spawn a new thread at the specified address
	threadId spawnThread(address addr);

	/// kill the specified thread
	void killThread(threadId tid);

	/// get the number of threads currently being executed
	uInt getNumThreads();

	/// get the current address of the specified
	/// thread
	/**
	 *	returns g_max_uInt if thread does not exist
	 */
	address getThreadAddress(threadId tid);

	/// test if the process has terminated
	/*
	*	a thread is considered terminated if it
	*	has no threads or has no program or the
	*	program size is 0
	*/
	virtual bool isTerminated();

protected:
	/// initalise dead process
	process();

	/// pointer to the program being executed
	program* m_pProgram;

	/// the number of active threads
	uInt m_threadCnt;

	/// the id of the process
	processId m_id;

private:
};





#endif


