

#ifndef __program_h__
#define __program_h__



#include "common.h"



/// holds the data for a program
class program
	: public std::vector<char>
{
public:
	program() {}
	~program() {}

	/// generate a random program from the specified seed
	/**
	 *	fills the currently allocated data with
	 *	random data
	 */
	void generate(uInt seed);

	/// returns the data at the specified address
	/**
	 *	the memory is assumed to wrap round
	 *	and so all addresses are valid unless the
	 *	program is of length 0, in which case no
	 *	address is valid.
	 */
	template <typename T>
	T read(address addr)
	{
		ASSERT(size()>0,"cannot read from 0 length program");

		// get the size
		address size_ = static_cast<address>(size());

		// make sure the address is in range
		addr%=size_;

		// check if the data overruns the end of the program
		if(addr+sizeof(T)<size_)
		{
			// if not then its simple
			// return the data at the address
			return *reinterpret_cast<T*>( &( (*this)[addr] ) );
		}

		// otherwise we need to  read part of the data
		// from the end of the program and part of it
		// from the start

		// temporaray variable to hold the return value
		T data_;

		// calculate the number of bytes of data before the split
		uInt split = size_-addr;

		// copy the first part of the data into the temporary
		// variable
		memcpy(&data_, &((*this)[addr]), split);

		// copy the second part of the data into the temporary
		// variable
		memcpy(reinterpret_cast<char*>((&data_))+split, &((*this)[0]), sizeof(T)-split);

		// return the temporary variable
		return data_;
	}

	/// writes the data to the specified address
	/**
	 *	the memory is assumed to wrap round
	 *	and so all addresses are valid unless the
	 *	program is of length 0, in which case no
	 *	address is valid.
	 */
	template <typename T>
	void write(address addr, T val)
	{
		ASSERT(size()>0,"cannot write to 0 length program");
		// get the size
		address size_ = static_cast<address>(size());

		// make sure the address is in range
		addr%=size_;

		// check if the data overruns the end of the program
		if(addr+sizeof(T)<size_)
		{
			// if not then its simple
			// copy the data to the address
			*reinterpret_cast<T*>( &( (*this)[addr] ) ) = val;
		} else
		{

			// otherwise we need to write part of the data
			// to the end of the program and part of it
			// to the beginning

			// calculate the number of bytes of data before the split
			uInt split = size_-addr;

			// copy the first part of the data into the temporary
			// variable
			memcpy(&((*this)[addr]), &val, split);

			// copy the second part of the data into the temporary
			// variable
			memcpy(&((*this)[0]), reinterpret_cast<char*>((&val))+split, sizeof(T)-split);
		}
	}

protected:
private:
};



#endif

